"""Test data loading and image fetching."""
import math

import numpy as np
import pytest

from mosaiks.fetch.images import fetch_image_crop, fetch_image_crop_from_stac_id
from mosaiks.fetch.stacs import fetch_stac_item_from_id


@pytest.fixture(scope="module")
def image_crop(config_dict: dict):
    """Test image crop."""
    lon, lat, id = (
        80.99266676800818,
        20.696269834517118,
        "LC08_L2SP_143046_20151208_02_T1",
    )
    stac_items = fetch_stac_item_from_id([id])

    return fetch_image_crop(
        lon,
        lat,
        stac_items,
        config_dict["image_width"],
        config_dict["image_bands"],
        config_dict["image_resolution"],
    )


@pytest.fixture(scope="module")
def image_crop_from_stac_id(config_dict: dict):
    """Stac Item fetched from stac ID."""
    lon, lat, id = (
        80.99266676800818,
        20.696269834517118,
        ["LC08_L2SP_143046_20151208_02_T1"],
    )
    return fetch_image_crop_from_stac_id(
        id,
        lon,
        lat,
        config_dict["image_width"],
        config_dict["image_bands"],
        config_dict["image_resolution"],
        config_dict["image_dtype"],
        config_dict["image_composite_method"],
        True,
        "planetary-compute",
    )


@pytest.fixture(scope="module")
def image_crop_from_nans(config_dict: dict):
    """Test image crop."""
    lon, lat, id = (np.nan, np.nan, None)
    stac_items = fetch_stac_item_from_id([id])

    return fetch_image_crop(
        lon,
        lat,
        stac_items,
        config_dict["image_width"],
        config_dict["image_bands"],
        config_dict["image_resolution"],
    )


@pytest.mark.parametrize(
    "test_image_crop",
    [
        pytest.lazy_fixture("image_crop"),
        pytest.lazy_fixture("image_crop_from_stac_id"),
        pytest.lazy_fixture("image_crop_from_nans"),
    ],
)
def test_image_crop_shape(test_image_crop: np.ndarray, config_dict: dict):
    image_size = math.ceil(
        config_dict["image_width"] / config_dict["image_resolution"] + 1
    )
    assert test_image_crop.shape == (
        len(config_dict["image_bands"]),
        image_size,
        image_size,
    )


@pytest.mark.parametrize(
    "test_image_crop",
    [pytest.lazy_fixture("image_crop"), pytest.lazy_fixture("image_crop_from_stac_id")],
)
def test_if_image_crop_is_normalised(test_image_crop: np.ndarray):
    assert np.nanmin(test_image_crop) >= 0 and np.nanmax(test_image_crop) <= 1


def test_if_image_crop_for_null_ids_are_nans(image_crop_from_nans: np.ndarray):
    assert ~np.isfinite(image_crop_from_nans).all()



import numpy as np
import torch

__all__ = ["create_features_from_image_array"]


def create_features_from_image_array(
    dataloader: torch.utils.data.DataLoader,
    n_features: int,
    model: torch.nn.Module,
    device: str,
    min_image_edge: str,
) -> np.ndarray:
    """
    Create features for images fetched through a dataloader.

    Parameters:
    -----------
    dataloader: A dataloader object that lazily yields images.
    n_features: The number of features to extract from each image.
    model: A model that extracts features from images.
    device: The device to run the model on.
    min_image_edge: The minimum edge length of an image to extract features from.

    Returns:
    --------
    features_array: An array with n_features columns.
    """

    n_points = len(dataloader.dataset)
    features_array = np.full([n_points, n_features], np.nan, dtype=float)
    model.eval().to(device)

    for i, image in enumerate(dataloader):
        if image is not None:
            # in case of single-band image, force-add a band dimension
            if len(image.shape) == 2:
                image = image.unsqueeze(0)

            if (image.shape[1] >= min_image_edge) and (
                image.shape[2] >= min_image_edge
            ):
                features_array[i] = featurize(image, model, device)
            else:
                # logging.warn("Image crop too small")
                pass
        else:
            # logging.warn("No image found")
            pass
    return features_array


def featurize(image: torch.Tensor, model: torch.nn.Module, device: str):
    """Helper method for running an image patch through a model.

    Parameters:
    -----------
    image: A tensor of shape (BANDS, X, Y) containing an image patch.
    model: A model that extracts features from images.
    device: The device to run the model on.

    Returns:
    --------
    feats: np.ndarray
        An array of shape (1, n_features) containing the extracted features.
    """
    image = image.to(device)
    feats = model(image).cpu().unsqueeze(0).numpy()
    return feats



"""Internal functions for checking inputs to pipeline functions."""

import numpy as np

valid_satellite_names = ["sentinel-2-l2a", "landsat-8-c2-l2"]
valid_stac_api_names = ["planetary-compute", "earth"]


def check_latitudes_and_longitudes(latitudes: list, longitudes: list) -> None:
    """Check that latitudes and longitudes are of equal length."""
    if len(latitudes) != len(longitudes):
        raise ValueError(
            f"Length of latitudes ({len(latitudes)}) must be equal to length of"
            f" longitudes ({len(longitudes)})"
        )

    # check that all latitudes and longitudes are numbers
    if not all(isinstance(lat, (int, float)) for lat in latitudes):
        raise ValueError("Latitudes must all be numbers.")
    if not all(isinstance(lon, (int, float)) for lon in longitudes):
        raise ValueError("Longitudes must all be numbers.")

    # check that there are no null values in latitudes and longitudes
    if np.isnan(latitudes).any() or np.isnan(longitudes).any():
        raise ValueError("Latitudes/longitudes cannot contain null values")


def check_satellite_name(satellite_name: str) -> None:
    """Check that satellite name is valid."""
    if satellite_name not in valid_satellite_names:
        raise ValueError(
            f"Satellite name must be one of {list(valid_satellite_names)},\
            not {satellite_name}"
        )


def check_stac_api_name(stac_api_name: str) -> None:
    """Check that stac api name is valid."""
    if stac_api_name not in valid_stac_api_names:
        raise ValueError(
            f"STAC api must be one of {list(valid_stac_api_names)},\
            not {stac_api_name}"
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

# Disabling the benchmarking feature causes cuDNN to deterministically
# select an algorithm, possibly at the cost of reduced performance.
# https://pytorch.org/docs/stable/notes/randomness.html
torch.backends.cudnn.benchmark = False


__all__ = ["RCF"]


class RCF(nn.Module):
    """
    A model for extracting Random Convolution Features (RCF) from input imagery.

    Parameters:
    -----------
    num_features: The number of features to extract from each image.
        NB: this should be an even number, since the features are produced by
        generating random filters, and concatenating the positive and negative
        convolutions from the filters.
    kernel_size: The size of the convolutional kernel.
    num_input_channels: The number of bands in the satellite image.
    random_seed_for_filters: The random seed to use when generating the filters.
        Defaults to 768.

    """

    def __init__(
        self,
        num_features: int = 1000,
        kernel_size: int = 3,
        num_input_channels: int = 6,
        random_seed_for_filters: int = 768,
    ):
        super().__init__()
        assert num_features % 2 == 0, "Please enter an even number of features."

        # Applies a 2D convolution over an input image composed of several input planes.
        self.conv1 = nn.Conv2d(
            num_input_channels,
            num_features // 2,
            kernel_size=kernel_size,
            stride=1,
            padding=0,
            dilation=1,
            bias=True,
        )

        # Set random state to initialise filters same way every time
        torch.manual_seed(random_seed_for_filters)

        # Fill the input Tensor with random gaussian noise and bias of -1
        nn.init.normal_(self.conv1.weight, mean=0.0, std=1.0)
        nn.init.constant_(self.conv1.bias, -1.0)

        # Explicitly freeze convolutional weights and bias
        self.conv1.weight.requires_grad_(False)
        self.conv1.bias.requires_grad_(False)

    def forward(self, x: torch.tensor) -> torch.tensor:
        """
        Parameters
        -----------
        x: A tensor of shape (BANDS, X, Y) containing an image patch.

        Returns
        --------
        A tensor containing the features.
        """

        assert x.shape[1:] > self.conv1.kernel_size, "Image too small for kernel size"
        x1a = F.relu(self.conv1(x))
        x1b = F.relu(-self.conv1(x))

        x1a = F.adaptive_avg_pool2d(x1a, (1, 1)).squeeze()
        x1b = F.adaptive_avg_pool2d(x1b, (1, 1)).squeeze()

        if len(x1a.shape) == 1:
            return torch.cat((x1a, x1b), dim=0)
        elif len(x1a.shape) >= 2:
            return torch.cat((x1a, x1b), dim=1)



"""Test dask code."""
from os import listdir
from pathlib import Path
from shutil import rmtree

import pandas as pd
import pytest

from mosaiks import utils as utl
from mosaiks.featurize import RCF
from mosaiks.pipeline.parallel import (
    _run_batched_pipeline,
    get_local_dask_cluster_and_client,
)


# -----Test local dask client-----
@pytest.mark.slow
def test__run_batched_pipeline(
    sample_test_data: pd.DataFrame,
    config_dict: dict,
):
    points_gdf = utl.df_w_latlons_to_gdf(sample_test_data)

    model = RCF(
        config_dict["n_mosaiks_features"],
        config_dict["mosaiks_kernel_size"],
        len(config_dict["image_bands"]),
    )
    columns = ["feature_%d" % i for i in range(config_dict["n_mosaiks_features"])]
    folder_path = Path("tests/data/test_output_batch_delayed/")
    folder_path.mkdir(parents=True, exist_ok=True)
    cluster, client = get_local_dask_cluster_and_client(2, 2)

    _run_batched_pipeline(
        points_gdf,
        client,
        model,
        satellite_name=config_dict["satellite_name"],
        image_resolution=config_dict["image_resolution"],
        image_dtype=config_dict["image_dtype"],
        image_bands=config_dict["image_bands"],
        image_width=config_dict["image_width"],
        min_image_edge=config_dict["min_image_edge"],
        sort_points_by_hilbert_distance=config_dict["sort_points_by_hilbert_distance"],
        datetime=config_dict["datetime"],
        image_composite_method=config_dict["image_composite_method"],
        stac_api_name=config_dict["stac_api_name"],
        num_features=config_dict["n_mosaiks_features"],
        device=config_dict["model_device"],
        n_concurrent_tasks=config_dict["dask_n_concurrent_tasks"],
        chunksize=config_dict["dask_chunksize"],
        col_names=columns,
        temp_folderpath=folder_path,
    )

    num_files = len(listdir(folder_path))
    rmtree(folder_path)
    client.close()
    cluster.close()

    assert num_files == 2



import logging
import shutil
from datetime import datetime as dt
from pathlib import Path
from typing import List, Optional

import dask.delayed
import dask_geopandas as dask_gpd
import geopandas as gpd
import numpy as np
import pandas as pd
import torch.nn as nn
from dask.distributed import Client, LocalCluster, as_completed

import mosaiks.utils as utl
from mosaiks.pipeline.standard import run_pipeline

__all__ = [
    "run_parallel_pipeline",
    "get_local_dask_cluster_and_client",
]


def run_parallel_pipeline(
    points_gdf: gpd.GeoDataFrame,
    model: nn.Module,
    satellite_name: str,
    image_resolution: int,
    image_dtype: str,
    image_bands: List[str],
    image_width: int,
    min_image_edge: int,
    datetime: str or list[str] or callable,
    image_composite_method: str,
    stac_api_name: str,
    n_mosaiks_features: int,
    model_device: str,
    mosaiks_col_names: list,
    n_concurrent_tasks: int,
    chunksize: int,
    client: Optional[Client],
    n_workers: Optional[int],
    threads_per_worker: Optional[int],
    sort_points_by_hilbert_distance: bool,
) -> pd.DataFrame:
    """
    For a given DataFrame of coordinate points, this function runs the `run_pipeline()`
    function on batches of datapoints in parallel using Dask.

    Parameters:
    -----------
    points_gdf : GeoDataFrame of points to be featurized.
    model: PyTorch model to be used for featurization.
    satellite_name: name of the satellite to use. Options are "landsat-8-c2-l2" or "sentinel-2-l2a".
    image_resolution: resolution of the satellite images in meters. Set depending on satellite.
    image_dtype: data type of the satellite images. Suggested "int16". All options - "int16", "int32", and "float"
    image_bands: list of bands to use for the satellite images. Suggested ["SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B6", "SR_B7"]. For options, read the satellite docs
    image_width: Desired width of the image to be fetched (in meters). Suggested 3000m for landsat.
    min_image_edge: minimum image edge in meters. Suggested 1000.
    datetime : date/times for fetching satellite images. See STAC API docs for `pystac.Client.search`'s `datetime` parameter for more details
    image_composite_method: how to composite multiple images for same GPS location. Options are "least_cloudy" (pick least cloudy image) or "all" (get all images and average across them). Suggested "least_cloudy" for speed.
    stac_api_name: which STAC API to use. Options are "planetary-compute" or "earth-search". Suggested "planetary-compute".
    n_mosaiks_features: number of mosaiks features to generate. Suggested 1000-4000.
    model_device: compute device for mosaiks model. Options are "cpu" or "cuda".
    mosaiks_col_names: column names for the mosaiks features.
    n_concurrent_tasks: number of concurrent tasks to run in Dask. Suggested None, which sets the total number of tasks to number of threads.
    chunksize: number of datapoints per data partition in Dask. Suggested 500.
    client : Dask client. If None, create a local cluster and client.
    threads_per_worker : Number of threads per worker. If None, let Dask decide (uses all available threads per core).
    sort_points_by_hilbert_distance: Whether to sort points by Hilbert distance before partitioning them. Suggested True.

    Returns
    --------
    DataFrame
    """

    # create a temporary directory
    date_time_now = dt.now().strftime("%Y-%m-%d_%H-%M-%S")
    temp_dir = Path.cwd() / f"dask_{date_time_now}"
    temp_dir.mkdir(parents=True, exist_ok=True)
    logging.info(f"Temporary directory: {temp_dir}")

    try:
        # Create a temporary dask client if one not given
        if client is None:
            logging.info("Creating a local Dask cluster and client...")
            cluster, client = get_local_dask_cluster_and_client(
                n_workers=n_workers, threads_per_worker=threads_per_worker
            )
            temp_client = True
        else:
            temp_client = False

        failed_partitions = _run_batched_pipeline(
            points_gdf=points_gdf,
            client=client,
            model=model,
            satellite_name=satellite_name,
            image_resolution=image_resolution,
            image_dtype=image_dtype,
            image_bands=image_bands,
            image_width=image_width,
            min_image_edge=min_image_edge,
            sort_points_by_hilbert_distance=sort_points_by_hilbert_distance,
            datetime=datetime,
            image_composite_method=image_composite_method,
            stac_api_name=stac_api_name,
            num_features=n_mosaiks_features,
            device=model_device,
            n_concurrent_tasks=n_concurrent_tasks,
            chunksize=chunksize,
            col_names=mosaiks_col_names,
            temp_folderpath=temp_dir,
        )
        if failed_partitions:
            logging.warn(f"Failed partitions: {failed_partitions}.")

        # IMPORTANT: Close dask client
        if temp_client:
            client.close()
            cluster.close()

        # Load checkpoint files and combine
        logging.info("Loading and combining checkpoint files...")
        checkpoint_filenames = utl.get_filtered_filenames(
            folder_path=temp_dir, prefix="df_"
        )
        combined_df = utl.load_and_combine_dataframes(
            folder_path=temp_dir, filenames=checkpoint_filenames
        )

    finally:
        # Delete temporary directory
        shutil.rmtree(temp_dir)

    return combined_df


def _run_batched_pipeline(
    points_gdf: gpd.GeoDataFrame,
    client: Client,
    model: nn.Module,
    satellite_name: str,
    image_resolution: int,
    image_dtype: str,
    image_bands: list[str],
    image_width: int,
    min_image_edge: int,
    sort_points_by_hilbert_distance: bool,
    datetime: str or list[str] or callable,
    image_composite_method: bool,
    stac_api_name: str,
    num_features: int,
    device: str,
    n_concurrent_tasks: int,
    chunksize: int,
    col_names: list[str],
    temp_folderpath: str,
    partition_ids: list[int] = None,
) -> list[int]:
    """
    Run partitions in batches and save the result for each partition
    to a parquet file.

    Parameters
    -----------
    points_gdf : GeoDataFrame of points to be featurized.
    client : Dask client.
    model : PyTorch model to be used for featurization.
    satellite_name : Name of satellite to be used for featurization.
    image_resolution : Resolution of satellite images to be used for featurization.
    image_dtype : Data type of satellite images to be used for featurization.
    image_bands : List of satellite image bands to be used for featurization.
    image_width : Desired width of the image to be fetched (in meters).
    min_image_edge : Minimum image edge size.
    sort_points_by_hilbert_distance: Whether to sort points by Hilbert distance before partitioning them. Suggested True.
    datetime: date/times for fetching satellite images. Same as datetime parameter in pystac.Client.search.
    image_composite_method : Mosaic composite to be used for featurization.
    stac_api_name : Name of STAC API to be used for satellite image search.
    num_features : number of mosaiks features.
    device : Device to be used for featurization.
    col_names : List of column names to be used for saving the features.
    temp_folderpath : Path to folder where features will be saved.
    partition_ids : Optional. List of partition indexes to be used for featurization.

    Returns
    --------
    list of failed partition ids
    """
    # make delayed gdf partitions
    dask_gdf = _get_dask_gdf(
        points_gdf,
        chunksize,
        sort_points_by_hilbert_distance,
    )
    partitions = dask_gdf.to_delayed()

    # only use the partitions specified by partition_ids if given
    if partition_ids is None:
        n_partitions = len(partitions)
        partition_ids = list(range(n_partitions))
    else:
        partitions = [partitions[i] for i in partition_ids]
        n_partitions = len(partitions)

    # if n_concurrent_tasks is not specified, use all available threads
    if n_concurrent_tasks is None:
        n_concurrent_tasks = sum(client.nthreads().values())
    # if there are less partitions to run than concurrent tasks, run all partitions
    n_concurrent_tasks = min(n_partitions, n_concurrent_tasks)

    logging.info(
        f"Running {n_partitions} partitions in batches of {n_concurrent_tasks} at a time."
    )

    failed_ids = []
    checkpoint_indices = list(np.arange(0, n_partitions, n_concurrent_tasks)) + [
        n_partitions
    ]
    for p_start_id, p_end_id in zip(checkpoint_indices[:-1], checkpoint_indices[1:]):
        now = dt.now().strftime("%d-%b %H:%M:%S")
        logging.info(f"{now} Running batch: {p_start_id} to {p_end_id - 1}")

        batch_indices = list(range(p_start_id, p_end_id))
        batch_p_ids = [partition_ids[i] for i in batch_indices]
        batch_partitions = [partitions[i] for i in batch_indices]

        failed_ids += _run_batch(
            partitions=batch_partitions,
            partition_ids=batch_p_ids,
            client=client,
            model=model,
            satellite_name=satellite_name,
            image_resolution=image_resolution,
            image_dtype=image_dtype,
            image_bands=image_bands,
            image_width=image_width,
            min_image_edge=min_image_edge,
            datetime=datetime,
            image_composite_method=image_composite_method,
            stac_api_name=stac_api_name,
            num_features=num_features,
            device=device,
            col_names=col_names,
            temp_folderpath=temp_folderpath,
        )

    return failed_ids


def _run_batch(
    partitions: list,
    partition_ids: list,
    client: Client,
    model: nn.Module,
    satellite_name: str,
    image_resolution: int,
    image_dtype: str,
    image_bands: list[str],
    image_width: int,
    min_image_edge: int,
    datetime: str or list[str] or callable,
    image_composite_method: str,
    stac_api_name: str,
    num_features: int,
    device: str,
    col_names: list,
    temp_folderpath: str,
) -> list[int]:
    """
    Run a batch of partitions and save the result for each partition to a parquet file.

    Parameters
    ----------
    partitions :List of dataframes to process.
    partition_ids : List containing IDs corresponding to the partitions passed (to be
        used for naming saved files and reference in case of failure).
    client : Dask client.
    model : PyTorch model to be used for featurization.
    satellite_name : Name of satellite to be used for featurization.
    image_resolution : Resolution of satellite images to be used for featurization.
    image_dtype : Data type of satellite images to be used for featurization.
    image_bands : List of satellite image bands to be used for featurization.
    image_width : Desired width of the image to be fetched (in meters).
    min_image_edge : Minimum image edge size.
    datetime : date/times for fetching satellite images. See STAC API docs for `pystac.Client.search`'s `datetime` parameter for more details
    image_composite_method : Mosaic composite to be used for featurization.
    stac_api_name : Name of STAC API to be used for satellite image search.
    num_features : number of mosaiks features.
    device : Device to be used for featurization.
    col_names : List of column names to be used for the output dataframe.
    temp_folderpath : Path to folder where features will be saved.

    Returns
    -------
    failed_ids : List of partition labels that failed to be featurized.
    """

    failed_ids = []
    delayed_tasks = []

    # collect delayed tasks
    for partition_id, partition in zip(partition_ids, partitions):
        str_id = str(partition_id).zfill(3)  # makes 1 into '001'

        # this can be swapped for delayed_pipeline(...)
        delayed_task = dask.delayed(run_pipeline)(
            points_gdf=partition,
            model=model,
            satellite_name=satellite_name,
            image_resolution=image_resolution,
            image_dtype=image_dtype,
            image_bands=image_bands,
            image_width=image_width,
            min_image_edge=min_image_edge,
            datetime=datetime,
            image_composite_method=image_composite_method,
            stac_api_name=stac_api_name,
            num_features=num_features,
            device=device,
            col_names=col_names,
            output_filepath=temp_folderpath / f"df_{str_id}.parquet",
            dask_key_name=f"features_{str_id}",
            return_df=False,
        )
        delayed_tasks.append(delayed_task)

    # compute delayed tasks
    futures = client.compute(delayed_tasks)

    # check for errors
    for completed_future in as_completed(futures):
        if completed_future.status == "error":
            f_key = completed_future.key
            partition_id = int(f_key.split("features_")[1])
            failed_ids.append(partition_id)

    return failed_ids


def get_local_dask_cluster_and_client(
    n_workers: Optional[int] = None,
    threads_per_worker: Optional[int] = None,
) -> tuple:
    """
    Get a local dask cluster and client.

    Parameters
    -----------
    threads_per_worker : Number of threads per worker. If None, let Dask decide (uses all available threads per core).
    dask_sort_points_by_hilbert_distance: Whether to sort points by Hilbert distance before partitioning them. Suggested True.

    Returns
    --------
    Dask cluster and client.
    """

    cluster = LocalCluster(
        n_workers=n_workers,
        threads_per_worker=threads_per_worker,
        silence_logs=logging.ERROR,
    )
    client = Client(cluster)

    # get number of workers and threads actually used by the client
    logging.info(
        f"\n\nDask client created with "
        f"{len(client.scheduler_info()['workers'])} workers and "
        f"{sum(client.nthreads().values())} total threads.\n"
        f"Dashboard link: {client.dashboard_link}\n\n"
    )

    return cluster, client


def _get_dask_gdf(
    points_gdf: gpd.GeoDataFrame,
    chunksize: int,
    sort_points_by_hilbert_distance: bool = True,
) -> dask_gpd.GeoDataFrame:
    """
    Split the gdf up by the given chunksize. To be used to create Dask Delayed jobs.

    Parameters
    ----------
    points_dgdf : A GeoDataFrame with a column named "geometry" containing shapely
        Point objects.
    chunksize : The number of points per partition to use creating the Dask
        GeoDataFrame.
    sort_points_by_hilbert_distance : Whether to sort the points by their Hilbert distance before splitting the dataframe into partitions.

    Returns
    -------
    points_dgdf: Dask GeoDataFrame split into partitions of size `chunksize`.
    """

    if sort_points_by_hilbert_distance:
        points_gdf = _sort_points_by_hilbert_distance(points_gdf)

    points_dgdf = dask_gpd.from_geopandas(
        points_gdf,
        chunksize=chunksize,
        sort=False,
    )

    logging.info(
        f"Created {points_dgdf.npartitions} partitions of {chunksize} points each."
    )

    return points_dgdf


def _sort_points_by_hilbert_distance(points_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Sort the points in the GeoDataFrame by their Hilbert distance."""

    ddf = dask_gpd.from_geopandas(points_gdf, npartitions=1)
    hilbert_distance = ddf.hilbert_distance().compute()
    points_gdf["hilbert_distance"] = hilbert_distance
    points_gdf = points_gdf.sort_values("hilbert_distance")

    return points_gdf



"""Test stac item fetching."""
import geopandas as gpd
import numpy as np
import pytest
from pystac.item import Item

import mosaiks.utils as utl
from mosaiks.fetch.stacs import fetch_image_refs, fetch_stac_item_from_id, get_stac_api


# -----Tests for stac item fetching-----
@pytest.fixture(scope="module")
def test_points_with_stac_with_all_composite(sample_test_data: gpd.GeoDataFrame):
    """Sample test data with stac items."""
    points_gdf = utl.df_w_latlons_to_gdf(sample_test_data)
    return fetch_image_refs(
        points_gdf,
        satellite_name="landsat-8-c2-l2",
        datetime=["2013-04-01", "2014-03-31"],
        stac_api_name="planetary-compute",
        image_composite_method="all",
    )


@pytest.fixture(scope="module")
def test_points_with_stac_with_least_cloudy_composite(
    sample_test_data: gpd.GeoDataFrame,
):
    """Sample test data with stac items."""
    points_gdf = utl.df_w_latlons_to_gdf(sample_test_data)
    return fetch_image_refs(
        points_gdf,
        satellite_name="landsat-8-c2-l2",
        datetime=["2013-04-01", "2014-03-31"],
        stac_api_name="planetary-compute",
        image_composite_method="least_cloudy",
    )


@pytest.mark.parametrize(
    "test_points_with_stac",
    [
        pytest.lazy_fixture("test_points_with_stac_with_all_composite"),
        pytest.lazy_fixture("test_points_with_stac_with_least_cloudy_composite"),
    ],
)
def test_if_stac_items_are_added_to_test_df(test_points_with_stac: gpd.GeoDataFrame):
    assert "stac_item" in test_points_with_stac.columns and type(
        test_points_with_stac["stac_item"].tolist()[0] == Item
    )


@pytest.mark.parametrize(
    "test_points_with_stac",
    [
        pytest.lazy_fixture("test_points_with_stac_with_all_composite"),
        pytest.lazy_fixture("test_points_with_stac_with_least_cloudy_composite"),
    ],
)
def test_if_df_with_stac_items_has_correct_shape(
    test_points_with_stac: gpd.GeoDataFrame, sample_test_data: gpd.GeoDataFrame
):
    assert len(test_points_with_stac) == len(sample_test_data)


# -----Tests for stac item fetching with null data-----
@pytest.fixture(scope="module")
def test_points_with_stac_null(sample_test_null_data: gpd.GeoDataFrame):
    """Sample test data with stac items."""
    points_gdf = utl.df_w_latlons_to_gdf(sample_test_null_data)
    return fetch_image_refs(
        points_gdf,
        satellite_name="landsat-8-c2-l2",
        datetime=["2013-04-01", "2014-03-31"],
        stac_api_name="planetary-compute",
        image_composite_method="least_cloudy",
    )


def test_if_stac_items_are_added_to_test_null_df(
    test_points_with_stac_null: gpd.GeoDataFrame,
):
    assert "stac_item" in test_points_with_stac_null.columns


def test_if_null_df_with_stac_items_has_correct_shape(
    test_points_with_stac_null: gpd.GeoDataFrame, sample_test_data: gpd.GeoDataFrame
):
    assert len(test_points_with_stac_null) == len(sample_test_data)


def test_if_stac_items_from_test_null_df_are_empty(
    test_points_with_stac_null: gpd.GeoDataFrame,
):
    assert test_points_with_stac_null["stac_item"].isnull().all()


# -----Tests for stac item fetch from stac ID-----
@pytest.fixture(scope="module")
def stac_item_from_stac_id():
    """Stac Item fetched from stac ID."""
    id = ["LC08_L2SP_143046_20151208_02_T1"]
    return fetch_stac_item_from_id(id, "planetary-compute")


def test_if_stac_item_is_returned_from_id(stac_item_from_stac_id: Item):
    assert type(stac_item_from_stac_id[0]) == Item


def test_if_stac_item_list_has_correct_shape(stac_item_from_stac_id: Item):
    assert len(stac_item_from_stac_id) >= 1


# -----Tests for stac item fetch for null stac ID-----
@pytest.fixture(scope="module")
def stac_item_from_null_stac_id():
    """Stac Item fetched from stac ID."""
    id = [None]
    return fetch_stac_item_from_id(id, "planetary-compute")


def test_if_null_stac_item_is_returned_from_null_id(stac_item_from_null_stac_id: Item):
    assert stac_item_from_null_stac_id[0] is None


def test_if_null_stac_item_list_has_correct_shape(stac_item_from_null_stac_id: Item):
    assert len(stac_item_from_null_stac_id) == 1



import logging
from pathlib import Path
from typing import List

import geopandas as gpd
import pandas as pd
import torch.nn

import mosaiks.utils as utl
from mosaiks.featurize import create_features_from_image_array
from mosaiks.fetch import create_data_loader, fetch_image_refs


def run_pipeline(
    points_gdf: gpd.GeoDataFrame,
    model: torch.nn.Module,
    satellite_name: str,
    image_resolution: int,
    image_dtype: str,
    image_bands: list,
    image_width: int,
    min_image_edge: int,
    datetime: str or List[str] or callable,
    image_composite_method: str,
    stac_api_name: str,
    num_features: int,
    device: str,
    col_names: list,
    output_filepath: Path = None,
    return_df: bool = True,
) -> pd.DataFrame:  # or None
    """
    For a given DataFrame of coordinate points, this function runs the necessary
    functions and optionally saves resulting mosaiks features to file.

    Parameters
    -----------
    points_gdf : GeoDataFrame of points to be featurized.
    model: PyTorch model to be used for featurization.
    satellite_name: name of the satellite to use. Options are "landsat-8-c2-l2" or "sentinel-2-l2a".
    image_resolution: resolution of the satellite images in meters. Set depending on satellite.
    image_dtype: data type of the satellite images. Suggested "int16". All options - "int16", "int32", and "float"
    image_bands: list of bands to use for the satellite images. Suggested ["SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B6", "SR_B7"]. For options, read the satellite docs
    image_width: Desired width of the image to be fetched (in meters). Suggested 3000m for landsat.
    min_image_edge: minimum image edge in meters. Suggested 1000.
    datetime : date/times for fetching satellite images. See STAC API docs for `pystac.Client.search`'s `datetime` parameter for more details
    image_composite_method: how to composite multiple images for same GPS location. Options are "least_cloudy" (pick least cloudy image) or "all" (get all images and average across them). Suggested "least_cloudy" for speed.
    stac_api_name: which STAC API to use. Options are "planetary-compute" or "earth-search". Suggested "planetary-compute".
    num_features: number of mosaiks features to generate. Suggested 1000-4000.
    device: compute device for mosaiks model. Options are "cpu" or "cuda".
    col_names: column names for the mosaiks features.
    output_filepath : Path to file where features will be saved. Must have .csv or .parquet or .parquet.gzip format. Default is None.
    return_df : Whether to return the features as a DataFrame. Default is True.

    Returns
    --------
    None or DataFrame depending on the "return_df" parameter.
    """

    points_gdf_with_stac = fetch_image_refs(
        points_gdf=points_gdf,
        satellite_name=satellite_name,
        datetime=datetime,
        image_composite_method=image_composite_method,
        stac_api_name=stac_api_name,
    )

    data_loader = create_data_loader(
        points_gdf_with_stac=points_gdf_with_stac,
        image_bands=image_bands,
        image_resolution=image_resolution,
        image_dtype=image_dtype,
        image_width=image_width,
        image_composite_method=image_composite_method,
    )

    X_features = create_features_from_image_array(
        dataloader=data_loader,
        n_features=num_features,
        model=model,
        device=device,
        min_image_edge=min_image_edge,
    )

    df = utl.make_result_df(
        features=X_features,
        context_gdf=points_gdf_with_stac,
        mosaiks_col_names=col_names,
    )

    if output_filepath is not None:
        try:
            utl.save_dataframe(df=df, file_path=output_filepath)
        except Exception as e:
            logging.error(f"Failed to save dataframe to {output_filepath}.")
            logging.error(e)

    if return_df:
        return df



from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd
import planetary_computer
import pystac_client
import shapely
from pystac.item import Item
from pystac.item_collection import ItemCollection

__all__ = ["fetch_image_refs", "fetch_stac_item_from_id"]


def fetch_image_refs(
    points_gdf: gpd.GeoDataFrame,
    satellite_name: str,
    datetime: str,
    image_composite_method: str,
    stac_api_name: str,
) -> gpd.GeoDataFrame:
    """
    Find the STAC item(s) that overlap each point in the `points_gdf` GeoDataFrame. Returns a
    GeoDataFrame with STAC items as a new column.

    Parameters
    ----------
    points_gdf : A GeoDataFrame with a column named "geometry" containing shapely
        Point objects.
    satellite_name : Name of MPC-hosted satellite.
    datetime : date/times for fetching satellite images. See STAC API docs for `pystac.Client.search`'s `datetime` parameter for more details
    image_composite_method : how to composite multiple images for same GPS location.
    stac_api_name: The name of the STAC API to use.

    Returns
    -------
    points_gdf: A GeoDataFrame with a column named "stac_item" containing STAC
        items.
    """
    stac_api = get_stac_api(stac_api_name)

    points_gdf_with_stac = points_gdf.copy()

    # Check for NaNs in lat lons
    nan_mask = points_gdf_with_stac["Lat"].isna() + points_gdf_with_stac["Lon"].isna()
    points_gdf_with_stac["stac_item"] = None

    if not nan_mask.all():
        points_gdf_not_nan = points_gdf_with_stac.loc[~nan_mask].copy()
        points_union = shapely.geometry.mapping(points_gdf_not_nan.unary_union)

        search_results = stac_api.search(
            collections=[satellite_name],
            intersects=points_union,
            datetime=datetime,
            query={"eo:cloud_cover": {"lt": 10}},
            limit=500,  # this limit seems arbitrary
        )
        item_collection = search_results.item_collection()

        if len(item_collection) == 0:
            return points_gdf_with_stac

        else:
            # Convert ItemCollection to GeoDataFrame
            if satellite_name == "landsat-8-c2-l2":
                # For landsat: trim the shapes to fit proj:bbox
                stac_gdf = _get_trimmed_stac_shapes_gdf(item_collection)
            else:
                # For Sentinel there is no need - there is no "proj:bbox" parameter
                # which could cause STACK issues
                stac_gdf = gpd.GeoDataFrame.from_features(item_collection.to_dict())

            # add items as an extra column
            stac_gdf["stac_item"] = item_collection.items

            points_gdf_with_stac.loc[~nan_mask] = _add_overlapping_stac_items(
                gdf=points_gdf_not_nan,
                stac_gdf=stac_gdf,
                image_composite_method=image_composite_method,
            )

    return points_gdf_with_stac


def _get_trimmed_stac_shapes_gdf(item_collection: ItemCollection) -> gpd.GeoDataFrame:
    """
    To prevent the edge case where a point sits inside the STAC geometry
    but outwith the STAC proj:bbox shape (resulting in a dud xarray to be
    returned later), trim the STAC shapes to within the proj:bbox borders.

    Returns
    -------
    GeoDataFrame where each row is an Item and columns include cloud cover percentage
    and item shape trimmed to within proj:bbox.
    """

    rows_list = []
    for i, item in enumerate(item_collection):
        stac_crs = item.properties["proj:epsg"]

        # get STAC geometry
        stac_geom = shapely.geometry.shape(item.geometry)

        # convert proj:bbox to polygon
        x_min_p, y_min_p, x_max_p, y_max_p = item.properties["proj:bbox"]
        image_bbox = shapely.geometry.Polygon(
            [
                [x_min_p, y_min_p],
                [x_min_p, y_max_p],
                [x_max_p, y_max_p],
                [x_max_p, y_min_p],
            ]
        )
        # convert to EPSG:4326 (to match STAC geometry)
        image_bbox = (
            gpd.GeoSeries(image_bbox).set_crs(stac_crs).to_crs(4326).geometry[0]
        )

        # trim stac_geom to only what's inside bbox
        trimmed_geom = stac_geom.intersection(image_bbox)

        row_data = {
            "eo:cloud_cover": [item.properties["eo:cloud_cover"]],
            "geometry": [trimmed_geom],
        }

        row = gpd.GeoDataFrame(row_data, crs=4326)
        rows_list.append(row)

    return pd.concat(rows_list)


def _add_overlapping_stac_items(
    gdf: gpd.GeoDataFrame,
    stac_gdf: gpd.GeoDataFrame,
    image_composite_method: str,
) -> List[Item]:
    """
    Takes in geodataframe of points and a sorted dataframe of stac items and returns
    the point geodataframe with an additional column which lists the items that covers
    each point. For use in `fetch_image_refs`.
    """

    gdf = gdf.copy()

    if image_composite_method == "least_cloudy":
        stac_gdf = stac_gdf.sort_values(by="eo:cloud_cover")

    for index, row in gdf.iterrows():
        items_covering_point = stac_gdf[stac_gdf.covers(row.geometry)]
        if len(items_covering_point) == 0:
            gdf.at[index, "stac_item"] = None
        else:
            if image_composite_method in ["all", "least_cloudy"]:
                all_items = items_covering_point["stac_item"].tolist()
                gdf.at[index, "stac_item"] = all_items
            else:
                raise ValueError(
                    f"image_composite_method must be 'least_cloudy' or 'all', not {image_composite_method}"
                )

    return gdf


def get_stac_api(stac_api_name: str) -> pystac_client.Client:
    """Get a STAC API client for a given API name."""

    if stac_api_name == "planetary-compute":
        stac_api = pystac_client.Client.open(
            "https://planetarycomputer.microsoft.com/api/stac/v1",
            modifier=planetary_computer.sign_inplace,
        )
    elif stac_api_name == "earth":
        stac_api = pystac_client.Client.open(
            "https://earth-search.aws.element84.com/v0"
        )
    else:
        raise NotImplementedError(f"STAC api {stac_api_name} is not implemented")

    return stac_api


# for debugging


def fetch_stac_item_from_id(
    ids: list[str], stac_api_name: str = "planetary-compute"
) -> list[Item]:
    """
    For debugging.

    Fetches STAC items from a list of ids.

    Parameters
    ----------
    ids: List of STAC item ids.
    stac_api_name: Name of STAC API to use. Has to be supported by `get_stac_api`.

    Returns
    -------
    List of STAC items.
    """
    stac_api = get_stac_api(stac_api_name)
    nan_mask = np.array([x is None for x in ids])
    if np.all(nan_mask):
        return [None] * len(ids)
    elif np.any(~nan_mask):
        search_results = [None] * len(ids)
        for i, id in enumerate(ids):
            if id is not None:
                stac_item_list = list(stac_api.search(ids=id).items())
                if len(stac_item_list) > 0:
                    search_results[i] = stac_item_list[0]

    return search_results



import logging
import os
from typing import List, Optional

import pandas as pd
from dask.distributed import Client

import mosaiks.checks as checks
import mosaiks.utils as utl
from mosaiks.featurize import RCF
from mosaiks.pipeline.parallel import run_parallel_pipeline
from mosaiks.pipeline.standard import run_pipeline

# Rasterio variables
# See https://github.com/pangeo-data/cog-best-practices
RASTERIO_CONFIG = {
    "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
    "GDAL_MAX_RAW_BLOCK_CACHE_SIZE": "200000000",
    "GDAL_SWATH_SIZE": "200000000",
    "VSI_CURL_CACHE_SIZE": "200000000",
    "AWS_REQUEST_PAYER": "requester",
}


def get_features(
    latitudes: List[float],
    longitudes: List[float],
    datetime: str or List[str] or callable,
    satellite_name: str = "landsat-8-c2-l2",
    image_resolution: int = 30,
    image_bands: List[str] = ["SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B6", "SR_B7"],
    image_width: int = 3000,
    min_image_edge: int = 30,
    image_composite_method: str = "least_cloudy",
    image_dtype: str = "int16",
    stac_api_name: str = "planetary-compute",
    n_mosaiks_features: int = 4000,
    mosaiks_kernel_size: int = 3,
    mosaiks_random_seed_for_filters: int = 768,
    model_device: str = "cpu",
    parallelize: bool = False,
    dask_chunksize: int = 500,
    dask_client: Optional[Client] = None,
    dask_n_workers: Optional[int] = None,
    dask_threads_per_worker: Optional[int] = None,
    dask_n_concurrent_tasks: Optional[int] = None,
    dask_sort_points_by_hilbert_distance: bool = True,
    setup_rasterio_env: bool = True,
) -> pd.DataFrame:
    """
    For a given set of coordinate points, this function runs the necessary functions,
    optionally with Dask parallel processing, and returns a dataframe of MOSAIKS features.

    Parameters:
    -----------
    latitudes: list of latitudes
    longitudes: list of longitudes
    datetime: date/times for fetching satellite images. See the STAC API documentation
        (https://pystac-client.readthedocs.io/en/latest/api.html#pystac_client.Client)
        for `.search`'s `datetime` parameter for more details.
    satellite_name: name of the satellite to use. Options are "landsat-8-c2-l2" or "sentinel-2-l2a". Defaults to "landsat-8-c2-l2".
    image_resolution: resolution of the satellite images in meters. Defaults to 30.
    image_bands: list of bands to use for the satellite images. Defaults to ["SR_B2", "SR_B3", "SR_B4", "SR_B5", "SR_B6", "SR_B7"]. For options, read the satellite docs
    image_width: Desired width of the image to be fetched (in meters). Default 3000m.
    min_image_edge: minimum image edge in meters. Defaults to 1000.
    image_composite_method: how to composite multiple images for same GPS location. Options are "least_cloudy" (pick least cloudy image) or "all" (get all images and average across them). Defaults to "least_cloudy".
    image_dtype: data type of the satellite images. Defaults to "int16". All options - "int16", "int32", and "float"
    stac_api_name: which STAC API to use. Options are "planetary-compute" or "earth-search". Defaults to "planetary-compute".
    n_mosaiks_features: number of mosaiks features to generate. Defaults to 4000.
    mosaiks_kernel_size: kernel size for mosaiks filters. Defaults to 3.
    mosaiks_random_seed_for_filters: random seed for mosaiks filters. Defaults to 768.
    model_device: compute device for mosaiks model. Options are "cpu" or "cuda". Defaults to "cpu".
    parallelize: whether to use Dask parallel processing. Defaults to False.
    dask_chunksize: number of datapoints per data partition in Dask. Defaults to 500.
    dask_client : Premade Dask client to use. If None, we create a new LocalCluster based on n_workers and threads_per_worker.
    n_workers : Number of workers to use. If None, let Dask decide (uses all available cores).
    threads_per_worker : Number of threads per worker. If None, let Dask decide (uses all available threads per core).
    dask_n_concurrent_tasks: number of concurrent tasks to run in Dask. Defaults to None, which sets the total number of tasks to number of threads.
    dask_sort_points_by_hilbert_distance: Whether to sort points by Hilbert distance before partitioning them. Defaults to True.
    setup_rasterio_env: whether to set up rasterio environment variables. Defaults to True.

    Returns
    --------
    DataFrame with MOSAIKS features for each input coordinate point and the corresponding STAC items of satellite images used.
    """

    # Set up Rasterio
    if setup_rasterio_env:
        os.environ.update(RASTERIO_CONFIG)

    # Check inputs
    logging.info("Checking inputs...")
    checks.check_latitudes_and_longitudes(latitudes, longitudes)
    checks.check_satellite_name(satellite_name)
    checks.check_stac_api_name(stac_api_name)

    # Make points df
    logging.info("Formatting data and creating model...")
    points_df = pd.DataFrame({"Lat": latitudes, "Lon": longitudes})

    # Convert points to gdf
    points_gdf = utl.df_w_latlons_to_gdf(points_df)

    mosaiks_col_names = [f"mosaiks_{i}" for i in range(n_mosaiks_features)]

    # Create model
    model = RCF(
        num_features=n_mosaiks_features,
        kernel_size=mosaiks_kernel_size,
        num_input_channels=len(image_bands),
        random_seed_for_filters=mosaiks_random_seed_for_filters,
    )

    logging.info("Getting MOSAIKS features...")
    if not parallelize:
        return run_pipeline(
            points_gdf=points_gdf,
            model=model,
            satellite_name=satellite_name,
            image_resolution=image_resolution,
            image_dtype=image_dtype,
            image_bands=image_bands,
            image_width=image_width,
            min_image_edge=min_image_edge,
            datetime=datetime,
            image_composite_method=image_composite_method,
            stac_api_name=stac_api_name,
            num_features=n_mosaiks_features,
            device=model_device,
            col_names=mosaiks_col_names,
            output_filepath=None,
            return_df=True,
        )
    else:
        return run_parallel_pipeline(
            points_gdf=points_gdf,
            model=model,
            satellite_name=satellite_name,
            image_resolution=image_resolution,
            image_dtype=image_dtype,
            image_bands=image_bands,
            image_width=image_width,
            min_image_edge=min_image_edge,
            datetime=datetime,
            image_composite_method=image_composite_method,
            stac_api_name=stac_api_name,
            n_mosaiks_features=n_mosaiks_features,
            model_device=model_device,
            mosaiks_col_names=mosaiks_col_names,
            n_concurrent_tasks=dask_n_concurrent_tasks,
            client=dask_client,
            chunksize=dask_chunksize,
            n_workers=dask_n_workers,
            threads_per_worker=dask_threads_per_worker,
            sort_points_by_hilbert_distance=dask_sort_points_by_hilbert_distance,
        )



import os
from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd


def load_dataframe(file_path: str, **kwargs) -> pd.DataFrame:
    """
    Load file with tabular data (csv or parquet) as a pandas DataFrame.

    Parameters
    ----------
    file_path : The path to the file to load.
    **kwargs : Keyword arguments to pass to the pandas read function.
    """

    file_path = str(file_path)

    if file_path.endswith(".csv"):
        return pd.read_csv(file_path, **kwargs)
    elif file_path.endswith(".parquet") or file_path.endswith(".parquet.gzip"):
        return pd.read_parquet(file_path, **kwargs)
    else:
        raise ValueError("File extension not recognized")


def load_and_combine_dataframes(folder_path: str, filenames: List[str]) -> pd.DataFrame:
    """
    Given folder path and filenames, load multiple dataframes and combine
    them, sording by the index.
    """

    dfs = []
    for filename in filenames:
        df = load_dataframe(file_path=folder_path / filename)
        dfs.append(df)

    combined_df = pd.concat(dfs, axis=0)
    combined_df.sort_index(inplace=True)

    return combined_df


def save_dataframe(df: pd.DataFrame, file_path: str, **kwargs) -> None:
    """
    Save pandas dataframe to .csv, .parquet etc file based on extension.

    Parameters
    ----------
    df : The dataframe to save.
    file_path : If given, the path to the file to load.
    """
    file_path = str(file_path)

    if os.path.dirname(file_path) != "":
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

    if file_path.endswith(".csv"):
        return df.to_csv(file_path, **kwargs)
    elif file_path.endswith(".parquet") or file_path.endswith(".parquet.gzip"):
        return df.to_parquet(file_path, **kwargs)
    else:
        raise ValueError("File extension not recognized.")


def df_w_latlons_to_gdf(
    df: pd.DataFrame,
    lat_name: str = "Lat",
    lon_name: str = "Lon",
    crs: str = "EPSG:4326",
) -> gpd.GeoDataFrame:
    """
    Convert DataFrame to GeoDataFrame by creating geometries from lat-lon columns.

    Parameters
    ----------
    df : The DataFrame to convert.
    lat_name, lon_name : The names of the columns containing the latitude and longitude values.
        Default is 'Lat' and 'Lon'.
    crs : The coordinate reference system of the lat-lon columns.
    """
    latlon_point_geoms = gpd.points_from_xy(x=df[lon_name], y=df[lat_name])
    gdf = gpd.GeoDataFrame(df.copy(), geometry=latlon_point_geoms, crs=crs)
    return gdf


def get_filtered_filenames(folder_path: str, prefix: str = "df_") -> List[str]:
    """
    Get the paths to all files in a folder that start with a given prefix.

    Parameters
    ----------
    folder_path : The path to the folder.
    prefix : The prefix that the filenames must start with.
    """

    all_filenames = os.listdir(folder_path)
    filtered_filenames = [file for file in all_filenames if file.startswith(prefix)]
    return sorted(filtered_filenames)


def make_result_df(
    features: np.ndarray,
    context_gdf: gpd.GeoDataFrame,
    mosaiks_col_names: list[str],
) -> pd.DataFrame:
    """
    Takes the features array and returns a dataframe with mosaiks features in the
    columns, plus the stac_id, with the same index as the context GeoDataFrame.

    Parameters
    -----------
    features : Array of features.
    context_gdf : GeoDataFrame of context variables. Must have the same index size as
        the features array. Must also have a "stac_item" column which contains
        pystac.item.Item objects since the "stac_id" is always saved.
    mosaiks_col_names : List of column names to label the feature columns as.

    Returns
    --------
    DataFrame
    """
    # Make features dataframe
    features_df = pd.DataFrame(
        data=features, index=context_gdf.index, columns=mosaiks_col_names
    )

    # Add stac_id to features dataframe
    if isinstance(context_gdf["stac_item"].iloc[0], list):
        features_df["stac_id"] = context_gdf["stac_item"].map(
            lambda item_list: [
                item.id if item is not None else None for item in item_list
            ]
        )
    else:
        features_df["stac_id"] = context_gdf["stac_item"].map(
            lambda item: item.id if item is not None else None
        )
    return features_df



"""Tests for featurization functions."""
import pytest
import torch

from mosaiks.featurize import RCF


@pytest.fixture(scope="module")
def model():
    """Fixture for a random convolutional features model."""
    return RCF(
        num_features=666,
        kernel_size=5,
        num_input_channels=3,
        random_seed_for_filters=42,
    )


@pytest.fixture(scope="module")
def model2():
    """Fixture for a random convolutional features model."""
    return RCF(
        num_features=666,
        kernel_size=5,
        num_input_channels=3,
        random_seed_for_filters=42,
    )


def test_random_model_conv_weights_are_created(model):
    assert hasattr(model, "conv1")


def test_random_model_conv_weights_are_not_trainable(model):
    assert (
        model.conv1.weight.requires_grad == False
        and model.conv1.bias.requires_grad == False
    )


def test_random_model_conv_weights_are_always_initialised_to_same_values(model, model2):
    assert (model.conv1.weight == model2.conv1.weight).all().item()


def test_random_conv_features_are_created(model):
    x1 = torch.empty(3, 10, 10)
    out1 = model(x1)
    assert out1.dim() == 1 and out1.shape[0] == 666


def test_random_conv_features_are_created_given_nan_image(model):
    x2 = torch.ones(3, 10, 10) * torch.nan
    out2 = model(x2)
    assert out2.dim() == 1 and out2.shape[0] == 666


def test_conv_features_are_nan_given_nan_image(model):
    x2 = torch.ones(3, 10, 10) * torch.nan
    out2 = model(x2)
    assert torch.isnan(out2).all().item()



import logging
import math
import warnings
from typing import List

import geopandas as gpd
import numpy as np
import pyproj
import stackstac
import torch
from pystac.item import Item
from torch.utils.data import DataLoader, Dataset

from mosaiks.fetch.stacs import fetch_stac_item_from_id

# temp warning suppress for stackstac 0.4.4's  UserWarning @ stackstac/prepare.py:364
# "pd.to_datetime() - UserWarning: The argument 'infer_datetime_format'"
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = [
    "create_data_loader",
    "fetch_image_crop_from_stac_id",
    "fetch_image_crop",
]


def fetch_image_crop(
    lon: float,
    lat: float,
    stac_items: list[Item],
    image_width: int,
    bands: List[str],
    resolution: int,
    dtype: str = "int16",
    image_composite_method: str = "least_cloudy",
    normalise: bool = True,
) -> np.array:
    """
    Fetches a crop of satellite imagery referenced by the given STAC item(s),
    centered around the given point and with the given image_width.

    If multiple STAC items are given, the median composite of the images is returned.

    Parameters
    ----------
    lon : Longitude of the centerpoint to fetch imagery for
    lat : Latitude of the centerpoint to fetch imagery for
    stac_items : list of STAC Items to fetch imagery for.
    image_width : Desired width of the image to be fetched (in meters).
        A buffer of image_width / 2 will be taken around the centerpoint.
    bands : List of bands to fetch
    resolution : Resolution of the image to fetch
    dtype : Data type of the image to fetch. Defaults to "int16".
        NOTE - np.uint8 results in loss of signal in the features
        and np.uint16 is not supported by PyTorch.
    image_composite_method : The type of composite to make if multiple images are given.
        If "least_cloudy"", take the least cloudy non-0 image. If "all", take a median
        composite of all images. Defaults to "least_cloudy"
    normalise : Whether to normalise the image. Defaults to True.

    Returns
    -------
    image : numpy array of shape (C, H, W)
    """
    if stac_items is None or all(x is None for x in stac_items):
        size = (
            len(bands),
            math.ceil(image_width / resolution + 1),
            math.ceil(image_width / resolution + 1),
        )
        return np.ones(size) * np.nan

    # Stac item must always be a list
    assert isinstance(stac_items, list)

    # calculate crop bounds
    # use the projection of the first non-None stac item
    stac_items_not_none = [item for item in stac_items if item is not None]
    crs = stac_items_not_none[0].properties["proj:epsg"]

    proj_latlon_to_stac = pyproj.Transformer.from_crs(4326, crs, always_xy=True)
    buffer_distance = image_width // 2 + 1
    x_utm, y_utm = proj_latlon_to_stac.transform(lon, lat)
    x_min, x_max = x_utm - buffer_distance, x_utm + buffer_distance
    y_min, y_max = y_utm - buffer_distance, y_utm + buffer_distance

    # remove the time dimension
    if image_composite_method == "all":
        # for a composite over all images, take median pixel over time
        xarray = stackstac.stack(
            stac_items_not_none,
            assets=bands,
            resolution=resolution,
            rescale=True,
            dtype=dtype,
            epsg=crs,  # set common projection for all images
            bounds=[x_min, y_min, x_max, y_max],
            fill_value=0,
        )
        image = xarray.median(dim="time").values

    elif image_composite_method == "least_cloudy":
        # for least cloudy, take the first non zero image
        for i, item in enumerate(stac_items_not_none):
            xarray = stackstac.stack(
                item,
                assets=bands,
                resolution=resolution,
                rescale=True,
                dtype=dtype,
                bounds=[x_min, y_min, x_max, y_max],
                fill_value=0,
            )
            image = xarray.values
            if len(image.shape) > 3:
                image = image.squeeze(0)

            # if image is not all zeros, break and use this image
            if ~np.all(image == 0.0):
                break
            else:
                # check next image if there are any stac items left
                if i < len(stac_items_not_none) - 1:
                    pass
                else:
                    logging.warning(
                        f"All images in the stack are zero for point {lon}, {lat}"
                    )
                    return np.ones_like(image) * np.nan

    if normalise:
        image = _minmax_normalize_image(image)

    return image


def _minmax_normalize_image(image: np.array) -> np.array:
    img_min, img_max = image.min(), image.max()
    return (image - img_min) / (img_max - img_min)


def create_data_loader(
    points_gdf_with_stac: gpd.GeoDataFrame,
    image_bands: List[str],
    image_resolution: int,
    image_dtype: str,
    image_width: int,
    image_composite_method: str = "least_cloudy",
) -> DataLoader:
    """
    Creates a PyTorch DataLoader which returns cropped images based on the given
    coordinate points and associated STAC image references.

    Parameters
    ----------
    points_gdf_with_stac : A GeoDataFrame with the points we want to fetch imagery for
        alongside the STAC item references to pictures for each point. The stac_item
        entry for each row should be a list of stac items that cover that point.
    image_bands : The bands to use for the image crops
    image_resolution : The resolution to use for the image crops
    image_dtype : The data type to use for the image crops
    image_width : Desired width of the image to be fetched (in meters).
    image_composite_method : The type of composite to make if multiple images are given.
        Defaults to "least_cloudy".

    Returns
    -------
    data_loader : A PyTorch DataLoader which returns cropped images as tensors
    """

    dataset = CustomDataset(
        latitudes=points_gdf_with_stac["Lat"].to_numpy(),
        longitudes=points_gdf_with_stac["Lon"].to_numpy(),
        stac_items=points_gdf_with_stac["stac_item"].tolist(),
        image_width=image_width,
        bands=image_bands,
        resolution=image_resolution,
        dtype=image_dtype,
        image_composite_method=image_composite_method,
    )
    data_loader = DataLoader(dataset, batch_size=None)

    return data_loader


class CustomDataset(Dataset):
    def __init__(
        self,
        latitudes: np.array,
        longitudes: np.array,
        stac_items: List[Item],
        image_width: int,
        bands: List[str],
        resolution: int,
        dtype: str = "int16",
        image_composite_method: str = "least_cloudy",
    ) -> None:
        """
        Parameters
        ----------
        latitudes : Array of latitudes to sample from
        longitudes : Array of longitudes to sample from
        stac_items : List of STAC items to sample from
        image_width : Desired width of the image to be fetched (in meters).
        bands : List of bands to sample
        resolution : Resolution of the image to sample
        dtype : Data type of the image to sample. Defaults to "int16".
            NOTE - np.uint8 results in loss of signal in the features
            and np.uint16 is not supported by PyTorch.
        image_composite_method : The type of composite to make if multiple images are given.
        """

        self.latitudes = latitudes
        self.longitudes = longitudes
        self.stac_items = stac_items
        self.image_width = image_width
        self.bands = bands
        self.resolution = resolution
        self.dtype = dtype
        self.image_composite_method = image_composite_method

    def __len__(self):
        """Returns the number of points in the dataset"""

        return len(self.latitudes)

    def __getitem__(self, idx: int) -> torch.Tensor:
        """
        Parameters
        ----------
        idx :Index of the point to get imagery for

        Returns
        -------
        out_image : Image tensor of shape (C, H, W)
        """

        lat = self.latitudes[idx]
        lon = self.longitudes[idx]
        stac_items = self.stac_items[idx]

        try:
            image = fetch_image_crop(
                lon=lon,
                lat=lat,
                stac_items=stac_items,
                image_width=self.image_width,
                bands=self.bands,
                resolution=self.resolution,
                dtype=self.dtype,
                image_composite_method=self.image_composite_method,
                normalise=True,
            )
            torch_image = torch.from_numpy(image).float()

            # if all 0s, replace with NaNs
            if torch.all(torch_image == 0):
                torch_image = np.ones_like(torch_image) * np.nan
            return torch_image

        except Exception as e:
            logging.warn(f"Skipping {idx}: {e}")
            return None


# for debugging


def fetch_image_crop_from_stac_id(
    stac_id: str,
    lon: float,
    lat: float,
    image_width: int,
    bands: List[str],
    resolution: int,
    dtype: str = "int16",
    image_composite_method: str = "least_cloudy",
    normalise: bool = True,
    stac_api_name: str = "planetary-compute",
) -> np.array:
    """
    Note: This function is necessary since STAC Items cannot be directly saved to file
    alongside the features in `.parquet` format, and so only their ID can be saved.
    This function uses the ID to first fetch the correct STAC Items, runs these
    through fetch_image_crop(), and optionally displays the image.

    Takes a stac_id (or list of stac_ids), lat, lon, and satellite
    parameters and returns and displays a cropped image. This image is fetched using the
    same process as when fetching for featurization.

    If multiple STAC items are given, the median composite of the images is returned.

    Parameters
    ----------
    stac_id : The STAC ID of the image to fetch
    lon : Longitude of the centerpoint to fetch imagery for
    lat : Latitude of the centerpoint to fetch imagery for
    image_width : Desired width of the image to be fetched (in meters).
    bands : The satellite image bands to fetch
    resolution : The resolution of the image to fetch
    dtype : The data type of the image to fetch. Defaults to "int16".
    image_composite_method : The type of composite to make if multiple images are given.
    normalise : Whether to normalise the image. Defaults to True.
    stac_api_name : The name of the STAC API to use. Defaults to "planetary-compute".

    Returns
    -------
    image_crop : A numpy array of the cropped image
    """

    if not isinstance(stac_id, list):
        stac_id = [stac_id]

    stac_items = fetch_stac_item_from_id(stac_id, stac_api_name)
    if len(stac_items) == 0:
        logging.warn("No STAC items found.")
        return None

    else:
        image_crop = fetch_image_crop(
            lon=lon,
            lat=lat,
            stac_items=stac_items,
            image_width=image_width,
            bands=bands,
            resolution=resolution,
            dtype=dtype,
            image_composite_method=image_composite_method,
            normalise=normalise,
        )

        return image_crop
